"""
This module is only available when running inside of maya.  It exposes
the  register_beforeOpen_callback()  command, which will run before maya
will open a file and is used to run the scanner if the file has an
ma extension.
"""
import maya.api.OpenMaya as api2
import maya.cmds as cmds
import manik.semantic.api
import manik.scan


_STATUS = {
    "kBeforeOpenCheck": 0,
    "kBeforeImportCheck": 0,
    "kBeforeCreateReferenceCheck": 0,
}


def ma_precheck(mFileObj, _cdata):
    """
    Maya kBeforeOpenCheck callback to validate
    MA files
    """

    # early out on non-ma files
    filename = mFileObj.resolvedFullName()
    if not filename.lower().endswith(".ma"):
        return False

    scanner = manik.scan.MAScanner.from_module(manik.semantic.api)
    result = scanner.scan_file(mFileObj.resolvedFullName())

    if result is not None:
        warning_string, offending = result

        # prettify the command string
        def expand_flags():
            for k, v in offending.flags.items():
                yield k
                if isinstance(v, list):
                    for item in v:
                        yield str(item)
                else:
                    yield str(v)

        def expand_args():
            for arg in offending.args:
                if isinstance(arg, list):
                    for a in arg:
                        yield str(a)
                else:
                    yield str(arg)

        mel_text = " ".join(
            [offending.command, " ".join(expand_flags()), " ".join(expand_args())]
        )

        # user confirm
        OPEN_UNSAFE = "Open unsafe file"
        DO_NOT_OPEN = "Do not open"

        open_anyway = cmds.confirmDialog(
            icon="critical",
            backgroundColor=(0.35, 0.18, 0.18),
            t="Suspicious content in MA file",
            m="This MA file may contain insecure data:\n\nThreat:{}\n\nLine: {}\n\nText:\n{}".format(
                warning_string, offending.lineno, mel_text
            ),
            b=[OPEN_UNSAFE, DO_NOT_OPEN],
            defaultButton=DO_NOT_OPEN,
            cancelButton=DO_NOT_OPEN,
            dismissString=DO_NOT_OPEN,
        )
        return open_anyway == OPEN_UNSAFE
    else:
        return True  # Expectiation is that we have no issues in the report: File is safe


def register_callbacks():
    """
    registers a kBeforeOpenCheck so Maya will not
    open files without the security check.

    If the callback has already been registered, does nothing.
    """

    if not _STATUS.get("kBeforeOpenCheck"):
        _STATUS["kBeforeOpenCheck"] = api2.MSceneMessage.addCheckFileCallback(
            api2.MSceneMessage.kBeforeOpenCheck, ma_precheck
        )

    if not _STATUS.get("kBeforeImportCheck"):
        _STATUS["kBeforeImportCheck"] = api2.MSceneMessage.addCheckFileCallback(
            api2.MSceneMessage.kBeforeImportCheck, ma_precheck
        )

    if not _STATUS.get("kBeforeCreateReferenceCheck"):
        _STATUS["kBeforeCreateReferenceCheck"] = api2.MSceneMessage.addCheckFileCallback(
            api2.MSceneMessage.kBeforeCreateReferenceCheck, ma_precheck
        )


def unregister_callbacks():
    for handler, event_id in _STATUS.copy().items():
        if event_id is None:
            continue

        try:
            api2.MMessage.removeCallback(event_id)
            _STATUS[handler] = 0
        except RuntimeError as e:
            cmds.warning(e)
