"""
manik.scan

This file implements the basic functionality of the scanner.  The actual analylitical work
is delegated to the api modules manik.semantic (for pyparsing style scans) or 
manik.lexical (for regex based scans).

To implement a new scanner, you need to provide two callables (they could be static functions,
instance methods, or functools.partials)

    parser
        A callable object which yields a stream of text chunks for analysis;
        for example, this could split the ma file on semicolons and return complete
        mel statements.     
    
    scanner  
        A callable object which parses the items yielded by the parser. If the
        text includes something suspicious it should return a tuple (warning, text)
        where warning is a message to the user and text is the offending content in the MA file.

To actually perform a scan, create an MAScanner object.

    my_scanner = MAScanner(my_parse_function, my_scan_function)

You can use the convenience method MAScanner.from_module() which will create them automatically 
from a module reference if the module provides the a `get_parser()` and a `get_scanner()` function.

An MAScanner object can scan a stream (a StringIO or other file-like object), a file, or a folder full 
of files. In all cases, a scan which finds no issues returns None so:

    problem = my_scanner.scan_file("path/to/file")
    if problem:
        print ("found " + problem)

or 
    problems = my_scanner.scan_folders("path/to/folder")
    if problems:
        for file, problem in problems.items():
            print (file, ":" problem)

"""
import time
import os
import logging
import sys


logger = logging.getLogger("manik.main")


PY2 = sys.version_info[0] < 3


def _elapsed(starttime):
    """
    convenience timer: returns in ms for under 2 seconds and in seconds above that
    """
    delta = time.time() - starttime
    if delta > 2:
        return "{} s".format(round(delta * 100) / 100.0)

    return "{} ms".format(round(delta * 1000))


class MAScanner(object):
    """
    Encapsulates 3 methods:

        scan_stream () takes a file-like object and scans it for
        problems, returning any potential problems or None

        scan_file () is a convenience wrapper that runs scan_stream()
        on a supplied disk path

        scan_folders() is a convenience wrapper that calls scan_file()
        on all the files in a folder and its children.

    the object needs to be configured with two callables -- a parsing function
    which splits the incoming text stream into pieces (probably, but not necessarily
    lines) and a scan function which looks at those pieces one at a time and
    returns any problematic issues
    """

    def __init__(self, parse_fn, scan_fn):
        self.parse = parse_fn
        self.scan = scan_fn

    def scan_stream(self, stream):
        """
        loop through the text chunks in <stream> looking for
        potential problems.  If found, return the problem
        and the offending text chunk.  If nothing is found,
        return None
        """
        logger.info("Scan started...")

        _t = time.time()
        ctr = 0
        result = None
        for chunk in self.parse(stream):
            ctr += 1
            chunk_result = self.scan(chunk)
            if chunk_result:
                result = chunk_result
                break

        logger.debug("    processed {} chunks".format(ctr))
        logger.info("...scan complete {}".format(_elapsed(_t)))

        if not result:
            logger.info("No issues found\n")
        else:
            warning_string, offending = result
            logger.warning("issue: {}".format(warning_string))
            logger.warning("in line {}".format(offending.lineno))
            logger.warning(offending)
            logger.warning("\n")

        return result

    def scan_file(self, filename):
        """
        Scans <filename> for problems.  If any items in the file
        present problems, return the problem as a tuple of
        (warning_string, original_text)

        If not, return None
        """
        logger.info("file: {}".format(filename))
        with open(filename, "rt") as filehandle:
            return self.scan_stream(filehandle)

    def scan_folders(self, *folders):
        """
        Applies scan_file() recursively to all the ma files in *folders.
        """
        _start = time.time()
        _ctr = 0
        results = {}
        for eachfolder in folders:
            for root, dirs, files in os.walk(eachfolder):
                ma_files = (f for f in files if f.lower().endswith(".ma"))
                abspaths = (os.path.abspath(os.path.join(root, ma)) for ma in ma_files)
                for eachpath in abspaths:
                    _ctr += 1
                    results[eachpath] = self.scan_file(eachpath)

        logger.info("scanned {} files in {}".format(_ctr, _elapsed(_start)))

        if results:
            errors = [v for v in results.values() if v]
            logger.warning("A total of {} issues were found".format(len(errors)))
        else:
            logger.info("no issues")

        return results

    @classmethod
    def from_module(cls, module):
        """
        creates a MAScanner object from a module reference which exposes "get_parser()" and a "get_scanner()"
        functions, where the parser yields a stream of text objects and the scanner returns possible
        security issues or None of a given piece of text appears safe
        """
        parser = module.get_parser()  # subdivides an input stream
        scanner = module.get_scanner()  # scans a chunk on input text for problems
        return cls(parser, scanner)
