from __future__ import print_function, division
from manik.semantic import ma_parser
from manik.semantic import ma_preprocessor_2
from manik.semantic import rules
import time
import logging
from collections import defaultdict, namedtuple
import pprint
import itertools

logger = logging.getLogger("manik.semantic.api")

"""
This file implements the generic manik api: the two functions

    get_parser()

and 

    get_scanner()

which return the parsing and scanning functions respectively. 

Here the scanner method iterates over a stream of text chunks
emitted by the ma_preprocessor (with associated line numbers)
and parses them as mel, looking for executable statements.  Any
executable code will be collected as a MelStatement namedtuple
in the form (command, args, flags, lineno).  Those objects will
then be passed to the rules defined in the rules module
and each obect will be checked against the rules to see
if it triggers an alert
"""


def get_parser():
    return parse_stream


def get_scanner():
    return scan


"""
These are the  objects we emit to the rules. The fields are:

    command -- usually, the executable mel statement. Sometimes a guard value (like "suspect string") 
    args    -- iterable of the arguments to the mel command 
    flags   -- dictionary of the flags to the mel command
    is_definition -- true if the statement is a mel procedure definition
    lineno  -- the line  number in the original file (for long, long  strings, its the first line #)
    context -- None for file statements, or a MelStatement if this expression was extracted from another MelStatement

"""
MelStatementBase = namedtuple("mel", "command args flags is_definition lineno context")


class MelStatement(MelStatementBase):
    def __repr__(self):
        full = (
            str(self.args)
            + " "
            + " ".join((k + " " + str(v)) for k, v in self.flags.items())
        )
        truncated = full[:128] + ("..." if len(full) > 64 else "")
        return "line {}{}: {} {}".format(
            self.lineno, "*" if self.context else "", self.command, truncated
        )


# common string tokens it's safe
# to ignore
_KNOWN_FALSE_POSITIVES = frozenset(
    (
        "persp",
        "top",
        "side",
        "front",
        "persp_depth",
        "top_depth",
        "side_depth",
        "front_depth",
        "persp_mask",
        "top_mask",
        "side_mask",
        "front_mask",
        "Maya",
        "map1",
        "Ambient+Diffuse",
        "Ambient",
    )
)

def normalize(pr, lineno, context):
    """
    given a parseResult, returns a MELSMelStatement which
    represetns a command, its arguments, and its flags. This
    decouples the rules from detailed knowledge of how the
    parseResults are organized
    """

    #logger.debug(">>>> %i%s-> %s", lineno,  "* " if context else " -", str(pr)[:78])

    # filter Nones
    if not pr:
        return

    if isinstance(pr, str):
        return

    # this flattens all of the parseResult data,
    # removing the keys
    if not isinstance(pr, list):
        pr = pr.asList()

    key = pr[0]
    if key == "None":
        return

    if key[0] == "/":
        return

    if key in _KNOWN_FALSE_POSITIVES:
        return

    # if key not in VALID_MEL_CMMANDS:
    #      return

    is_definition = "proc" in pr[:2]
    if is_definition:
        procname = pr[-2]
        return MelStatement(procname, [pr], {}, True, lineno, context)

    # in case we have a parenthetical expression, like
    # (LocalizedPanel("panel"))
    if "[" in key:
        pr = pr[0]
        key = str(pr[0])

    pr_len = len(pr)
    if pr_len == 1:
        return MelStatement(key, [], {}, is_definition, lineno, context)
    elif pr_len == 2:
        return MelStatement(key, pr[1], {}, is_definition, lineno, context)
    else:
        result_dict = {}
        for stub in pr[2:]:
            if stub:
                result_dict[stub[0]] = stub[1:]

        is_embed = result_dict.pop("-manik_embed", None) is not None
        if is_embed and not context:
            # somebody has snuck executable code into a file-level setAttr or similar
            pr[1].insert(0, key)
            key = "embedded"

        return MelStatement(key, pr[1], result_dict, is_definition, lineno, context)


def expand_statement(parseResult):
    """
    Filter a parse result, yielding executable statements including
    executable statements in arguments or flags.

    Note that this will turn some things that aren't really functions
    (eg, a string attribute that's just a name or similar) -- we pay
    a small tax for checking those but if an attacker is trying to
    stash executable code in an inert attribute this  will catch it.
    """

    try:
        proc = parseResult.get("m_proc")
    except AttributeError:
        return

    if proc:
        yield proc

    imp = parseResult.get("m_imp")
    if imp:
        yield imp

        # catch sneaky attempts to insert a backtick into the args
        # add a flag to the resulting command so the normalizer
        # can make it obvious this is a bad thing if we are
        # at the file level
        arguments = parseResult.get("m_arg")

        try:
            call = arguments.get("m_call") or arguments.get("m_func")
            if call:
                call.append(["-manik_embed"])
                yield call
        except AttributeError:
            pass

        flags = (f for f in imp[2:] if len(f) > 1 and hasattr(f, "get"))
        for f in flags:
            for child in expand_statement(f):
                child.append(["-manik_embed"])
                yield child

    ass = parseResult.get("m_assign")
    if ass:
        # yields any imperatives on the right side of an
        # assignment
        rhs = ass.get("m_imp")
        if rhs:
            yield rhs
            for child in expand_statement(rhs):
                yield child
        return

    func = parseResult.get("m_func")
    if func:
        yield func
        for arg in itertools.chain.from_iterable(func[1:]):
            for child in expand_statement(arg):
                yield child
        return

    call = parseResult.get("m_call")
    if call:
        yield call
        for arg in itertools.chain.from_iterable(call[1:]):
            for child in expand_statement(arg):
                yield child
        return


def parse_stream(filelike):
    """
    Parses <filelike>, an object which supports 'read()' -- for example
    as StringIO or a file handle.
    """
    _t = time.time()
    parser = ma_parser.get_parser()
    binary_parser = ma_parser.get_binary_parser()

    filtered = ma_preprocessor_2.prefilter_stream(filelike)

    stats = defaultdict(int)

    def traverse(source, lineno, context):
        # source is a text source to be scanned for parseResults
        # lineno is the linenumber from the original input
        # level is 0 for file-level text and 1 for text in string attributes
        raw = iter(parser.searchString(source))
        expanded = (expand_statement(s) for s in raw if s)
        include_subs = itertools.chain.from_iterable(expanded)
        normalized = (normalize(n, lineno, context) for n in include_subs)
        skip_nulls = (k for k in normalized if k)

        return skip_nulls

    try:
        context = None

        for root_level, embedded_level, lineno in filtered:

            logger.debug("\n %i :  %s | %s", lineno, root_level , embedded_level[:32])
            

            for result in traverse(root_level, lineno, None):

                if isinstance(result.command, list):
                    stats[result.command[0]] += 1
                else:
                    stats[result.command] += 1
                logger.debug(result)
                yield result
                if result.command == "setAttr":
                    context = result

            if root_level == '<UIConfig>':
                context = root_level

            for result in traverse(embedded_level, lineno, context):
                stats[result.command] += 1
                logger.debug(result)
                yield result

                
    except Exception as e:
        logger.exception(e)

    finally:
        logger.debug("found {0} executable statements".format(sum(stats.values())))
        logger.debug("command stats:\n{0}".format(pprint.pformat(dict(stats))))
        logger.info("stream completed in {0}".format(elapsed(_t)))


"""
SAFE is a list of commands we know we don't care about expanding-- 
things that can't reasonably be part of an attack unless

We need to use this sparingly: it's will cut down
on the cost of processing rules, but if we accidentally
identify a function as harmless and its not, we have
hidden it from the scanner.  

example innocent looking but unsafe: `modelEditor` has a 
callback script which could be hijacked by setting code
with the '-ec' flag, looking like one of the many annoying
uiConfiguration scripts

Note that a lot of setAttrs and Renames are already
filtered in the preprocessor as well.  Filtering the
setAttrs here does NOT filter string attributes which 
might carry code -- that's getting sent to the secondary stream

"""


_SAFE = frozenset(
    (
        "localizedPanelLabel",
        "sceneUIReplacement",
        "viewSet",
        "lambert1",
        "lambert2",
        "grid",
        "panelHistory",
        "select",
        "connectAttr",
        "currentUnit",
        "requires",
        "stereoCameraView",
        "findStartUpCamera"
    )
)


def is_safe_call(melstmt):
    """
    given a Melstatement, returns False
    if we regard it as 'harmless' (ie,
    it's not capable of triggering attack
    code in its native form).  This saves us
    running some number of rules against the
    statement.

    We need to use this sparingly: it's will cut down
    on the cost of processing rules, but if we accidentally
    identify a function as harmless and its not, we have
    hidden it from the scanner.

    Remember we're trusting there expander to pry out
    subordinate expressions (like a callable function argument)
    separately

    Be careful about including new items in the safe list:
    An example innocent looking but unsafe command is  `modelEditor`
    which has a callback script which could be hijacked for bad
    ends.  Only add a command to the safe list if (a) it's
    numerous and (b) it has not callbacks or executable flags
    """
    if melstmt.is_definition:
        return False

    return melstmt.command in _SAFE


def scan(stmt, *rulesets):
    """
    Applies the rules defined in *rulesets (or all known
    rules, if rulesets is not supplied) to the
    """

    if is_safe_call(stmt):
        return

    for each_scan in rules.get_rules(*rulesets):
        warning = each_scan(stmt)
        if warning:
            return warning, stmt


def elapsed(starttime):
    """
    convenience timer: returns in ms for under 2 seconds and in seconds above that
    """
    delta = time.time() - starttime

    if delta > 2:
        return "{0} s".format(round(delta * 100) / 100.0)
    else:
        return "{0} ms".format(round(delta * 1000))
