from logging import Logger
import logging
import re
import io
import time
import itertools
import ast
from unittest import runner
from .mel_reference import UI_CONF_BAD_FLAGS

logger = logging.getLogger("manik.semantic.preprocessor")

#DEBUG CODE REMOVE
from manik.scan import MAScanner


# RE's for quickly identifying potential trouble spots
_inlineable = re.compile('''(`)|(\()''')
_stringAtrr = re.compile('''setAttr \".*\" -type \"string\"''')
_common_commands = re.compile("(renam)|(conne)|(relat)|(paren)|(selec)|(fileI)|(setAt)|(addAtt)|(insta)")
_scary_callbacks = re.compile("(\.pram)|(\.poam)|(\.prlm)|(\.polm)|(\.prm)|(\.pom)|(\"\.a\")|(\.hc)|(\.homeCamera)|(\.expression)|(\"\.e\")|(\.internalExpression)|(\.preFurRenderMel)|(\.preMel)|(\.postMel)|(\.preRenderLayerMel)|(\.postRenderLayerMel)|(\.preRenderMel)|(\.postRenderMel)|(\.preFurRenderMel)|(\.postFurRenderMel)")
_confignode = re.compile("((scene)|(ui))ConfigurationScriptNode")
_viewset = re.compile('"viewSet -. %camera"')
_scriptCreate = re.compile("(createNode) (script)")

# stereoptyped things that always show up in the UIConfig which we don't need to parse
# these need to be syntactically airtitght so pay attention to anchors and string escapes
_scripted_panel = re.compile('^(shape|scripted|model|outliner|pose)Panel -edit -l \(localizedPanelLabel\(\".*\"\)\) -mbv .menusOkayInPanels  .panelName;$')         
_sceneUI = re.compile('\$panelName = `sceneUIReplacement -getNextPanel \".*\" \(localizedPanelLabel\(\".*\"\)\) `;$')
_panelname= re.compile('} }   \$panelName = `sceneUIReplacement -getNextScriptedPanel \".*\" \(localizedPanelLabel\(\".*\"\)\)')
_ifnot = re.compile('^if \(\"\" != \$panelName\) { \$label = `panel -q -label \$panelName`;$')
_sceneConfig = re.compile('^if \(!.useSceneConfig\) { panel -e -l .label .panelName;$')
_editorname = re.compile('^\s*\$editorName = (\$.*|\(\$panelName\+\".*\"\));$')
_selected = re.compile('^modelEditor -e -viewSelected 0 \$editorName;$')
_gpu = re.compile('^modelEditor -e -pluginObjects "gpuCacheDisplayFilter" 1 \$editorName;$')
_vars = re.compile('^string \$.*;$')
_comment= re.compile('^\W*//')
_funcSyntax = re.compile('\w+\(')
_safe_ui_config = (_comment, _scripted_panel, _sceneUI, _panelname, _ifnot, _sceneConfig, _editorname, _selected, _gpu, _vars)

# we can skip a lot of the common ui config junk if (a) it has no executable chars and (b) no bad flags
_bad_flag_set = set(UI_CONF_BAD_FLAGS["modelEditor"])
_bad_flag_set.add(UI_CONF_BAD_FLAGS['outlinerEditor'])
_suspect_uiconf_flags = re.compile("|".join ("(-{})".format(k) for k in _bad_flag_set))

# these functions form a generator chain. The more aggressive filtering should come early in the chain.

def yield_lines(raw):
    """
    yield out all of the lines in file-like object <raw>, with multi-line values 
    concatenated to formw single units.  

    The results are yielded as tuples of line number, text .  Whitespace is stripped
    """
    ctr = 0
    previous = []
    previous_line = 0
    for line in raw.readlines():
        ctr += 1
        line = line.strip()
        if not previous:
            if line.endswith(";") or line.startswith("//"):
                yield ctr, line
            else:
                if not previous:
                    previous_line = ctr 
                previous.append(line)

        elif not line.endswith(";"):
            previous.append(line)
        else:
            previous.append(line)
            yield previous_line,  " ".join(previous)
            del previous[:]
    # don't forget to clear the buffer if they didn't put a semi at the end of an inserted line!
    if previous_line:
        yield previous_line,  " ".join(previous)

            

def is_setattr_or_inline(incoming):
    """
    In a well-formed Mel file, we only care about setAttrs or items which could be
    executable.  So we filter out the common commands unless they include a possible
    inlined call.  This eliminates the vast majority of the file contents
    """
    for ctr, item in incoming:
        if item.startswith("//"):
            continue
        inline = _inlineable.search(item)
        if inline:
            # the inline trigger may indicate a multiline string assembly...
            # if so, expand it to make life easier on the parser
            if _stringAtrr.search(item) and item.endswith(");"):
                funcsig = _funcSyntax.search(item)

                # distinguish between func syntax and multiline assembly
                if funcsig and funcsig.start() < inline.start():
                    yield ctr, item, funcsig.start()    
                else:
                    multiline =  expand_multilines(item)
                    yield ctr, multiline, inline.start()
            else:
                # if our inline has a "(", we need to backtrack

                yield ctr, item, inline.start()
        elif not _common_commands.search(item[:6]):
            yield ctr, item, False
        elif _stringAtrr.search(item) and  _scary_callbacks.search(item[:32]):
            yield ctr, item, False


def filter_creates(incoming):
    """
    We only care are about scriptnodes or from 'createNode'
    """
    for ctr, item, inline in incoming:
        if inline:
            yield ctr, item, inline
            continue
        if item[0:3] != 'cre':
            if not _viewset.search(item):
                # suppress viewsSet callbacks if they are
                # exactly the default ones
                yield ctr, item, inline

        else:
            if _scriptCreate.search(item[:24]) and not _confignode.search(item):
                yield ctr, item, False




def defer_ui_config_node(incoming):
    """it's expensive to parse the UIConfig node, so only do it if there are no other hits..."""
    ui_line, ui_node = None, None

    for ctr, item, inline in incoming:

        if ui_node:
            yield ctr, item, inline, False
        elif 'Maya Mel UI Configuration File' in item:
            ui_line, ui_node = ctr, item
        else:
            yield ctr, item, inline, False
    
    if ui_node:
        merged = decompose_escaped(ui_node)
        for _, line in merged:
            is_safe = any(b.search(line) for b in _safe_ui_config)
            if not is_safe:
                yield  ui_line, line, False, True


def expand_multilines(item):
    """expand mel formatted multilines line ("str" + "str" + "str")"""    
    start = item.index("(") + 1
    trimmed = item[start: -2].replace('" + "', '')
    return  item[:start -1 ] + " " + trimmed 

def decompose_escaped(escaped_text):
    """
    this reconstitutes the UIConfiguration scriptnode
    """
    start = escaped_text.index('"string"') + 13
    unescaped = escaped_text[start -1:].replace("\\\\n", "\n")\
        .replace("\\\\t", "\t")\
        .replace("\\\\\"", '"')\
        .replace("\\n", "\n")\
        .replace("\\t", "\t")\
        .replace("\\\"",'"')

    
    class DummyFileLike(object):
        """
        mock a readlines interface so we can reuse yield_lines on a string
        """
        def __init__(self, stringbuffer):
            self.stringbuffer = stringbuffer

        def readlines(self):
            return  iter(self.stringbuffer.splitlines())
                

    for item in yield_lines(DummyFileLike(unescaped)):
        yield item



def prefilter_stream(stream):
    """
    Filters the file into a stream of easier-to-parse strings.

    The stream consists of tuples, the first item is the raw call
    (for example 'createNode') and the second item is the contents
    of any string attribute set by `setAttr` (for example the contents)
    of the uiConfiguration script nodes.  The second item is
    unescaped so other code does not need to worry about double quotes
    and so on.

    We separate these streams because executable code inline in the
    main file is always malicious where there is a lot of innocent
    code (such as the camera home commands) living in string attributes;
    forking them allows us to develop separate rules
    """
    raw = yield_lines(stream)

    all_sets =  is_setattr_or_inline(raw)
    only_sus = filter_creates(all_sets)
    ui_last = defer_ui_config_node(only_sus)

    for num, text, inline, is_ui in ui_last:
        if not is_ui:
            if not inline:
                yield text, '', num
            else:
                yield text, text[inline:], num
        else:          
            if text.count(";") > 1:
                for t in text.split(";"):
                    yield "<UIConfig>",t, num
            else:
                yield "<UIConfig>",text, num

                # I made an effort to trim out the 'safe" modelEditorCallbacks, which are numerous
                # but my logic did not catch all examples....
                '''
                if "`" in text or "(" in text:
                    yield "<UIConfig>",text, num
                elif not "modelEditor" in text and not "outlinerEditor" in text:
                    yield "<UIConfig>",text, num
                if _suspect_uiconf_flags.search(text):
                    yield "<UIConfig>",text, num
                '''


def get_parser():
    return prefilter_stream


def get_scanner():
    def dont_parse(item):
        logger.debug (item)
        return

    return dont_parse
