# signing.py - MA code signing
from __future__ import division
import os
import hashlib
import uuid
from collections import namedtuple


Signature = namedtuple('signature', 'hash mac version')
Hashresult = namedtuple('hashresult', 'hash signed filesize')

MANIK_SALT = b'3b9615aa811f86495771f40a5478802e6938c3d4'
SIG_LINE = b'// MANIK '
SIG_SIZE = 160
SIG_VERSION = 1000
CHUNK_SIZE = 16384


def hash_file(filehandle, secret):  
    """
    hashes the contents of the file in <filehandle> using the supplied secret

    returns a Hashresult object indicating the hash, whether or not the file has been signed previously, and the current file size
    """
    filehandle.seek(0, os.SEEK_END)
    size = filehandle.tell()

    filehandle.seek(-SIG_SIZE, os.SEEK_END)
    last_block = filehandle.read()

    is_signed = False
    if SIG_LINE in last_block:
        is_signed = True
        stop_at = size-SIG_SIZE
    else:
        stop_at = size

    filehandle.seek(0, os.SEEK_SET)
    file_hash = hashlib.sha512(bytes(secret))
    file_hash.update(MANIK_SALT)
    
    pointer = 0
    last_chunk_size = stop_at % CHUNK_SIZE
    while pointer < stop_at:
        next_chunk = filehandle.read(CHUNK_SIZE)
        pointer += CHUNK_SIZE
        if pointer > stop_at:
            next_chunk = next_chunk[:last_chunk_size]
        file_hash.update(next_chunk)
    return Hashresult(file_hash.hexdigest(), is_signed, size)


def get_signature(file_hash, mac):
    """
    returns the ascii bytes for a signature comment line given the 128-bit file hash and 48-byte MAC address
    """
    sig = "{line}{hash}0{mac:#016}0{version}\n".format(
        line=SIG_LINE.decode('utf-8'),
        hash=file_hash,
        mac=mac,
        version=SIG_VERSION,
    ).encode('ascii')
    return sig


def sign(filename, secret=0):
    """
    Sign the file using the supplied secret.  The last 160 bytes of the file will be
    a comment in the form :

        // MANIK [128 byte hash in hex form][16 character MAC address][4 character version number]
   
    """
    mac = uuid.getnode()

    with open(filename, 'a+b') as filehandle:
        file_hash, already_signed, size = hash_file(filehandle, secret)
        if already_signed:
            filehandle.truncate(size-SIG_SIZE)
        filehandle.seek(0, os.SEEK_END)

        sig = Signature(file_hash, mac, SIG_VERSION)
        sig_string = get_signature(sig.hash, sig.mac)

        filehandle.write(sig_string)

    return sig


def parse_signature(sig_bytes):
    """
    parses the signature comment <sig_bytes> and returns a Signature object containing the 
    file hash, mac address, and version number contained in the comment. 
    """
    if len(sig_bytes) != SIG_SIZE:
        raise ValueError("Signature must be {} characters.".format(SIG_SIZE))

    sig = sig_bytes.decode('ascii')
    crc = sig[9:137].encode('ascii')
    mac = int(sig[138:155])
    version = int(sig[155:159])

    return Signature(crc, mac, version)


def validate(filename, secret=0):
    """
    Validates filename against an expected secret
    """
    with open(filename, 'rb') as filehandle:
        filehandle.seek(-SIG_SIZE, os.SEEK_END)
        sig_block = filehandle.read()
        if SIG_LINE not in sig_block:
            print("file is unsigned")
            return -1

        file_sig = hash_file(filehandle, secret)

    got_hash = parse_signature(sig_block)
    return got_hash.hash == file_sig.hash.encode('ascii')


def get_manik_metadata(filename):
    """
    Returns a Signature object detailing the file hash, mac address, and version of the 
    signature in file <filename>, or None if the file is not signed
    """
    with open(filename, 'rb') as filehandle:
        filehandle.seek(-SIG_SIZE, os.SEEK_END)
        sig_block = filehandle.read()
        if SIG_LINE not in sig_block:
            return None
